"""
Blockchain network main class and its derivatives.

Nodeos api reference:
https://developers.eos.io/manuals/eos/latest/nodeos/plugins/chain_api_plugin/api-reference/index
"""

import typing
from urllib.parse import urljoin

import httpx
import pydantic

from pyntelope import exc
from pyntelope._version import __version__


class Net(pydantic.BaseModel):
    """
    The net hold the connection information with the blockchain network api.
    """  # NOQA: D200

    host: pydantic.AnyHttpUrl

    def _request(
        self,
        *,
        endpoint: str,
        payload: typing.Optional[dict] = dict(),
        verb: str = "POST",
    ):
        url = urljoin(self.host, endpoint)
        headers = {"user-agent": f"pyntelope/{__version__}"}

        try:
            resp = httpx.post(url, json=payload, headers=headers)
        except (
            httpx.TimeoutException,
            httpx.NetworkError,
            httpx.WriteError,
        ) as e:
            raise exc.ConnectionError(
                response=None, url=url, payload=payload, error=e
            )

        if resp.status_code > 299 and resp.status_code != 500:
            raise exc.ConnectionError(
                response=resp, url=url, payload=payload, error=None
            )

        return resp.json()

    def abi_bin_to_json(
        self, *, account_name: str, action: str, bytes: dict
    ) -> dict:
        endpoint = "/v1/chain/abi_bin_to_json"
        payload = dict(code=account_name, action=action, binargs=bytes.hex())
        data = self._request(endpoint=endpoint, payload=payload)
        return data["args"]

    def abi_json_to_bin(
        self, *, account_name: str, action: str, json: dict
    ) -> bytes:
        """
        Return a dict containing the serialized action data.

        https://developers.eos.io/manuals/eos/latest/nodeos/plugins/chain_api_plugin/api-reference/index#operation/abi_json_to_bin
        """
        endpoint = "/v1/chain/abi_json_to_bin"
        payload = dict(code=account_name, action=action, args=json)
        data = self._request(endpoint=endpoint, payload=payload)
        if "binargs" not in data:
            return data
        hex_ = data["binargs"]
        bytes_ = bytes.fromhex(hex_)
        return bytes_

    def get_info(self):
        endpoint = "/v1/chain/get_info"
        data = self._request(endpoint=endpoint)
        return data

    def get_account(self, *, account_name: str):
        """
        Return an account information.

        If no account is found, then raises an connection error
        https://developers.eos.io/manuals/eos/latest/nodeos/plugins/chain_api_plugin/api-reference/index#operation/get_account
        """
        endpoint = "/v1/chain/get_account"
        payload = dict(account_name=account_name)
        data = self._request(endpoint=endpoint, payload=payload)
        return data

    def get_abi(self, *, account_name: str):
        """
        Retrieve the ABI for a contract based on its account name.

        https://developers.eos.io/manuals/eos/latest/nodeos/plugins/chain_api_plugin/api-reference/index#operation/get_abi
        """
        endpoint = "/v1/chain/get_abi"
        payload = dict(account_name=account_name)
        data = self._request(endpoint=endpoint, payload=payload)
        if len(data) == 1:
            return None
        return data

    def get_block(self, *, block_num_or_id: str):
        """
        Return various details about a specific block on the blockchain.

        https://developers.eos.io/manuals/eos/latest/nodeos/plugins/chain_api_plugin/api-reference/index#operation/get_block
        """
        endpoint = "/v1/chain/get_block"
        payload = dict(block_num_or_id=block_num_or_id)
        data = self._request(endpoint=endpoint, payload=payload)
        return data

    def get_block_info(self, *, block_num: str):
        """
        Return a fixed-size smaller subset of the block data.

        Similar to get_block
        https://developers.eos.io/manuals/eos/latest/nodeos/plugins/chain_api_plugin/api-reference/index#operation/get_block_info
        """
        endpoint = "/v1/chain/get_block_info"
        payload = dict(block_num=block_num)
        data = self._request(endpoint=endpoint, payload=payload)
        return data

    def get_table_by_scope(
        self,
        code: str,
        table: str = None,
        lower_bound: str = None,
        upper_bound: str = None,
        limit: int = None,
        reverse: bool = None,
        show_payer: bool = None,
    ):
        """
        Return a dict with all tables and their scope.

        Similar to get_table_by_scope
        https://developers.eos.io/manuals/eos/latest/nodeos/plugins/chain_api_plugin/api-reference/index#operation/get_table_by_scope
        """
        endpoint = "/v1/chain/get_table_by_scope"
        payload = dict(
            code=code,
            table=table,
            lower_bound=lower_bound,
            upper_bound=upper_bound,
            limit=limit,
            reverse=reverse,
            show_payer=show_payer,
        )
        for k in list(payload.keys()):
            if payload[k] is None:
                del payload[k]
        data = self._request(endpoint=endpoint, payload=payload)
        return data

    def get_table_rows(
        self,
        code: str,
        table: str,
        scope: str,
        json: bool = True,
        index_position: str = None,
        key_type: str = None,
        encode_type: str = None,
        lower_bound: str = None,
        upper_bound: str = None,
        limit: int = None,
        reverse: int = None,
        show_payer: int = None,
    ):
        """
        Return a list with the rows in the table.

        Similar to get_table_rows
        https://developers.eos.io/manuals/eos/latest/nodeos/plugins/chain_api_plugin/api-reference/index#operation/get_table_rows

        Parameters:
        -----------
        json: bool = True
            Get the response as json
        """
        endpoint = "/v1/chain/get_table_rows"
        payload = dict(
            code=code,
            table=table,
            scope=scope,
            json=json,
            index_position=index_position,
            key_type=key_type,
            encode_type=encode_type,
            lower_bound=lower_bound,
            upper_bound=upper_bound,
            limit=limit,
            reverse=reverse,
            show_payer=show_payer,
        )
        for k in list(payload.keys()):
            if payload[k] is None:
                del payload[k]
        data = self._request(endpoint=endpoint, payload=payload)
        if "rows" in data:
            return data["rows"]
        else:
            return data

    def push_transaction(
        self,
        *,
        transaction: object,
        compression: bool = False,
        packed_context_free_data: str = "",
    ):
        """
        Send a transaction to the blockchain.

        https://developers.eos.io/manuals/eos/latest/nodeos/plugins/chain_api_plugin/api-reference/index#operation/push_transaction
        """
        endpoint = "/v1/chain/push_transaction"
        payload = dict(
            signatures=transaction.signatures,
            compression=compression,
            packed_context_free_data=packed_context_free_data,
            packed_trx=transaction.pack(),
        )
        data = self._request(endpoint=endpoint, payload=payload)
        return data


class WaxTestnet(Net):
    host: pydantic.HttpUrl = "https://testnet.waxsweden.org/"


class WaxMainnet(Net):
    host: pydantic.HttpUrl = "https://facings.waxpub.net"


class EosMainnet(Net):
    host: pydantic.HttpUrl = "https://api.eossweden.org"


class KylinTestnet(Net):
    host: pydantic.HttpUrl = "https://kylin.eossweden.org"


class Jungle3Testnet(Net):
    host: pydantic.HttpUrl = "https://jungle3.eossweden.org"


class TelosMainnet(Net):
    host: pydantic.HttpUrl = "https://telos.caleos.io/"


class TelosTestnet(Net):
    host: pydantic.HttpUrl = "https://testnet.telos.caleos.io"


class ProtonMainnet(Net):
    host: pydantic.HttpUrl = "https://proton.cryptolions.io"


class ProtonTestnet(Net):
    host: pydantic.HttpUrl = "https://testnet.protonchain.com"


class UosMainnet(Net):
    host: pydantic.HttpUrl = "https://uos.eosusa.news"


class FioMainnet(Net):
    host: pydantic.HttpUrl = "https://fio.cryptolions.io"


class Local(Net):
    host: pydantic.HttpUrl = "http://127.0.0.1:8888"


__all__ = [
    "Net",
    "EosMainnet",
    "KylinTestnet",
    "Jungle3Testnet",
    "TelosMainnet",
    "TelosTestnet",
    "ProtonMainnet",
    "ProtonTestnet",
    "UosMainnet",
    "FioMainnet",
    "WaxTestnet",
    "WaxMainnet",
    "Local",
]



"""Antelope data types."""

import calendar
import datetime as dt
import re
import struct
import sys
from abc import ABC, abstractmethod

import pydantic


class AntelopeType(pydantic.BaseModel, ABC):
    def __init__(self, *args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0:
            super().__init__(value=args[0])
        else:
            super().__init__(*args, **kwargs)

    @pydantic.validator("value", pre=True, check_fields=False)
    def check_if_same_type(cls, v):
        if type(v) is cls:
            return v.value
        return v

    @abstractmethod
    def __bytes__(self):
        """Convert instance to bytes."""

    @abstractmethod
    def from_bytes(self):
        """Create instance from bytes."""

    def __len__(self):
        """Lenght of value in bytes."""
        bytes_ = bytes(self)
        return len(bytes_)

    class Config:
        extra = "forbid"
        frozen = True


class UnixTimestamp(AntelopeType):
    """
    Serialize a datetime.

    Precision is in seconds
    Considers UTC time
    """

    value: dt.datetime

    @pydantic.validator("value")
    def remove_everything_bellow_seconds(cls, v):
        new_v = v.replace(microsecond=0)
        return new_v

    def __bytes__(self):
        unix_secs = calendar.timegm(self.value.timetuple())
        uint32_secs = Uint32(value=unix_secs)
        bytes_ = bytes(uint32_secs)
        return bytes_

    @classmethod
    def from_bytes(cls, bytes_):
        uint32_secs = Uint32.from_bytes(bytes_)
        datetime = dt.datetime.utcfromtimestamp(uint32_secs.value)
        return cls(value=datetime)


class Bool(AntelopeType):
    value: bool

    def __bytes__(self):
        return b"\x01" if self.value else b"\x00"

    @classmethod
    def from_bytes(cls, bytes_):
        return cls(value=int(bytes_[:1].hex(), 16))


class String(AntelopeType):
    value: str

    def __bytes__(self):
        bytes_ = self.value.encode("utf8")
        length = len(bytes_)
        bytes_ = bytes(Varuint32(value=length)) + bytes_
        return bytes_

    @pydantic.validator("value")
    def must_not_contain_multi_utf_char(cls, v):
        if len(v) < len(v.encode("utf8")):
            msg = (
                f'Input "{v}" has a multi-byte utf character in it, '
                "currently pyntelope does not support serialization of "
                "multi-byte utf characters."
            )
            raise ValueError(msg)
        return v

    @classmethod
    def from_bytes(cls, bytes_):
        size = Varuint32.from_bytes(bytes_)
        start = len(size)
        string_bytes = bytes_[start : start + size.value]  # NOQA: E203
        value = string_bytes.decode("utf8")
        return cls(value=value)


class Asset(AntelopeType):
    """
    Serialize a Asset.

    serializes an amount (can be a float value) and currency name together
    uses Symbol type to serialize percision and name of currency,
    uses Uint64 type to serialize amount
    amount and name are seperated by one space
    example: 50.100000 WAX
    """

    value: str

    def get_name(self):
        """
        Extract the name from a raw Asset string.

        example: "WAX" from Asset string "99.1000000 WAX"
        """
        stripped_value = self.value.strip()
        return stripped_value.split(" ")[1]

    def get_int_digits(self):
        """
        Extract the integer digits (digits before the decimal).

        from raw Asset string
        example: "99" from Asset string "99.1000000 WAX"
        """
        stripped_value = self.value.strip()
        pos = 0
        int_digits = ""

        # check for negative sign
        if stripped_value[pos] == "-":
            int_digits += "-"
            pos += 1

        curr_char = stripped_value[pos]

        # get amount value
        while (
            pos < len(stripped_value) and curr_char >= "0" and curr_char <= "9"
        ):
            int_digits += curr_char
            pos += 1
            curr_char = stripped_value[pos]

        return int_digits

    def get_frac_digits(self):
        """
        Extract the decimal digits as integers (digits after the decimal).

        example: "1000000" from Asset string "99.1000000 WAX"
        """
        stripped_value = self.value.strip()
        pos = 0
        precision = 0
        frac_digits = ""
        curr_char = 0

        if "." in stripped_value:
            pos = stripped_value.index(".") + 1
            curr_char = stripped_value[pos]
            while (
                pos < len(stripped_value)
                and curr_char >= "0"  # noqa: W503
                and curr_char <= "9"  # noqa: W503
            ):
                frac_digits += curr_char
                pos += 1
                curr_char = stripped_value[pos]
                precision += 1

        else:
            return ""

        return frac_digits

    def get_precision(self):
        """
        Get the precision (number of digits after decimal).

        example: "7" from Asset string "99.1000000 WAX"
        """
        return len(self.get_frac_digits())

    def __bytes__(self):

        amount = Uint64(int(self.get_int_digits() + self.get_frac_digits()))
        name = self.get_name()
        symbol = Symbol(str(self.get_precision()) + "," + name)

        amount_bytes = bytes(amount)
        symbol_bytes = bytes(symbol)

        return amount_bytes + symbol_bytes

    @classmethod
    def from_bytes(cls, bytes_):
        amount_bytes = bytes_[:8]  # get first 8 bytes
        asset_precision = bytes_[8]  # (amount of decimal places)
        amount = str(
            struct.unpack("<Q", amount_bytes)[0]
        )  # amount with decimal values (no decimal splitting yet)
        # get name (currency) from Symbol
        name = str(Symbol.from_bytes(bytes_[8:])).split(",")[1][:-1]
        if asset_precision == 0:
            value = amount + " " + name
        else:
            value = (
                amount[:-asset_precision]
                + "."  # noqa: W503
                + amount[asset_precision + 1 :]  # noqa: W503, E203
                + " "  # noqa: W503
                + name  # noqa: W503
            )  # combine everything and place decimal in correct position
        return cls(value=value)

    @pydantic.validator("value")
    def amount_must_be_in_the_valid_range(cls, v):
        value_list = str(v).strip().split(" ")
        if len(value_list) != 2:
            msg = (
                f'Input "{v}" must have exactly one space in between '
                "amount and name"
            )
            raise ValueError(msg)
        return v

    @pydantic.validator("value")
    def check_for_frac_digit_if_decimal_exists(cls, v):
        stripped_value = v.strip()
        if "." in stripped_value:
            pos = stripped_value.index(".") + 1
            curr_char = stripped_value[pos]
            if (
                pos < len(stripped_value)
                and curr_char >= "0"  # noqa: W503
                and curr_char <= "9"  # noqa: W503
            ):
                return v
            else:
                msg = (
                    "decimal provided but no fractional digits were provided."
                )
                raise ValueError(msg)
        return v

    @pydantic.validator("value", allow_reuse=True)
    def check_if_amount_is_valid(cls, v):
        stripped_value = v.strip()
        amount = float(stripped_value.split(" ")[0])
        if amount < 0 or amount > 18446744073709551615:
            msg = f'amount "{amount}" must be between 0 and ' "2^64 inclusive."
            raise ValueError(msg)
        return v

    @pydantic.validator("value", allow_reuse=True)
    def check_if_name_is_valid(cls, v):
        stripped_value = v.strip()
        name = stripped_value.split(" ")[1]
        match = re.search("^[A-Z]{1,7}$", name)
        if not match:
            msg = f'Input "{name}" must be A-Z and between 1 to 7 characters.'
            raise ValueError(msg)
        return v


class Symbol(AntelopeType):
    """
    Serialize a Symbol.

    serializes a percision and currency name together
    precision is used to indicate how many decimals there
    are in an Asset type amount
    precision and name are seperated by a commma
    example: 1,WAX
    """

    value: str

    @pydantic.validator("value", allow_reuse=True)
    def name_must_be_of_valid_length(cls, v):
        name = v.split(",")[1]
        match = re.search("^[A-Z]{1,7}$", name)
        if not match:
            msg = f'Input "{name}" must be A-Z and between 1 to 7 characters.'
            raise ValueError(msg)
        return v

    @pydantic.validator("value", allow_reuse=True)
    def precision_must_be_in_the_valid_range(cls, v):
        precision = int(v.split(",")[0])
        if precision < 0 or precision > 16:
            msg = (
                f'precision "{precision}" must be between 0 and '
                "16 inclusive."
            )
            raise ValueError(msg)
        return v

    def __bytes__(self):
        precision = int(self.value.split(",")[0])
        precision_bytes_ = struct.pack("<B", (precision & 0xFF))
        bytes_ = precision_bytes_
        name = self.value.split(",")[1]
        name_bytes_ = name.encode("utf8")
        bytes_ += name_bytes_
        leftover_byte_space = len(name) + 1
        while (
            leftover_byte_space < 8
        ):  # add null bytes in remaining empty space
            bytes_ += struct.pack("<B", 0)
            leftover_byte_space += 1
        return bytes_

    @classmethod
    def from_bytes(cls, bytes_):
        bytes_len = len(bytes_)
        precision = ""
        name = ""
        for i in range(bytes_len):
            if chr(bytes_[i]).isupper():
                precision = str(bytes_[0])
                name_bytes = bytes_[i:]  # name is all bytes after precision
                for k in range(1, len(name_bytes) + 1):
                    if not chr(bytes_[k]).isupper():
                        name_bytes = name_bytes[
                            : k - 1
                        ]  # name only goes up to the last upper case character
                name = name_bytes.decode("utf8")
                break

        value = precision + "," + name
        return cls(value=value)


class Bytes(AntelopeType):
    value: bytes

    def __bytes__(self):
        return self.value

    @classmethod
    def from_bytes(cls, bytes_):
        return cls(value=bytes_)


class Array(AntelopeType):
    values: tuple
    type_: type

    @pydantic.validator("type_")
    def must_be_subclass_of_antelope(cls, v):
        if not issubclass(v, AntelopeType):
            raise ValueError("Type must be subclass of AntelopeType")
        return v

    @pydantic.root_validator(pre=True)
    def all_must_satisfy_type_value(cls, all_values):
        type_ = all_values["type_"]
        values = all_values["values"]
        if len(values) >= 1:
            values = tuple(type_(v) for v in values)
        else:
            values = tuple()
        all_values["values"] = values
        return all_values

    def __bytes__(self):
        bytes_ = b""
        length = Varuint32(len(self.values))
        bytes_ += bytes(length)
        for value in self.values:
            bytes_ += bytes(value)
        return bytes_

    @classmethod
    def from_bytes(cls, bytes_, type_):
        length = Varuint32.from_bytes(bytes_)
        bytes_ = bytes_[len(length) :]  # NOQA: E203
        values = []
        for n in range(length.value):
            value = type_.from_bytes(bytes_)
            values.append(value.value)
            bytes_ = bytes_[len(value) :]  # NOQA: E203
        return cls(values=values, type_=type_)

    def __getitem__(self, index):
        return Array(values=self.values[index], type_=self.type_)


class Name(AntelopeType):
    # regex = has at least one "non-dot" char
    value: pydantic.constr(
        max_length=13,
        regex=r"^[\.a-z1-5]*[a-z1-5]+[\.a-z1-5]*$|^(?![\s\S])",  # NOQA: F722
    )

    def __eq__(self, other):
        """Equality diregards dots in names."""
        if type(other) != type(self):
            return False
        return self.value.replace(".", "") == other.value.replace(".", "")

    @pydantic.validator("value")
    def last_char_restriction(cls, v):
        if len(v) == 13:
            allowed = {"a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "."}
            if v[-1] not in allowed:
                msg = (
                    "When account name is 13 char long, last char must be "
                    f'in range a-z or ".". "{v[-1]}" found.'
                )
                raise ValueError(msg)
        return v

    def __bytes__(self):
        value_int = self.string_to_uint64(self.value)
        uint64 = Uint64(value=value_int)
        return bytes(uint64)

    @classmethod
    def from_bytes(cls, bytes_):
        uint64 = Uint64.from_bytes(bytes_)
        value_str = cls.uint64_to_string(uint64.value, strip_dots=True)
        return cls(value=value_str)

    @classmethod
    def char_to_symbol(cls, c):
        if c >= ord("a") and c <= ord("z"):
            return (c - ord("a")) + 6
        if c >= ord("1") and c <= ord("5"):
            return (c - ord("1")) + 1
        return 0

    @classmethod
    def string_to_uint64(cls, s):
        if len(s) > 13:
            raise Exception("invalid string length")
        name = 0
        i = 0
        while i < min(len(s), 12):
            name |= (cls.char_to_symbol(ord(s[i])) & 0x1F) << (
                64 - 5 * (i + 1)
            )
            i += 1
        if len(s) == 13:
            name |= cls.char_to_symbol(ord(s[12])) & 0x0F
        return name

    @classmethod
    def uint64_to_string(cls, n, strip_dots=False):
        charmap = ".12345abcdefghijklmnopqrstuvwxyz"
        s = bytearray(13 * b".")
        tmp = n
        for i in range(13):
            c = charmap[tmp & (0x0F if i == 0 else 0x1F)]
            s[12 - i] = ord(c)
            tmp >>= 4 if i == 0 else 5

        s = s.decode("utf8")
        if strip_dots:
            s = s.strip(".")
        return s


class Int8(AntelopeType):
    value: pydantic.conint(ge=-128, lt=128)

    def __bytes__(self):
        return struct.pack("<b", self.value)

    @classmethod
    def from_bytes(cls, bytes_):
        struct_tuple = struct.unpack("<b", bytes_[:1])
        value = struct_tuple[0]
        return cls(value=value)


class Uint8(AntelopeType):
    value: pydantic.conint(ge=0, lt=256)  # 2 ** 8

    def __bytes__(self):
        return struct.pack("<B", self.value)

    @classmethod
    def from_bytes(cls, bytes_):
        struct_tuple = struct.unpack("<B", bytes_[:1])
        value = struct_tuple[0]
        return cls(value=value)


class Uint16(AntelopeType):
    value: pydantic.conint(ge=0, lt=65536)  # 2 ** 16

    def __bytes__(self):
        return struct.pack("<H", self.value)

    @classmethod
    def from_bytes(cls, bytes_):
        struct_tuple = struct.unpack("<H", bytes_[:2])
        value = struct_tuple[0]
        return cls(value=value)


class Uint32(AntelopeType):
    value: pydantic.conint(ge=0, lt=4294967296)  # 2 ** 32

    def __bytes__(self):
        return struct.pack("<I", self.value)

    @classmethod
    def from_bytes(cls, bytes_):
        struct_tuple = struct.unpack("<I", bytes_[:4])
        value = struct_tuple[0]
        return cls(value=value)


class Uint64(AntelopeType):
    value: pydantic.conint(ge=0, lt=18446744073709551616)  # 2 ** 64

    def __bytes__(self):
        return struct.pack("<Q", self.value)

    @classmethod
    def from_bytes(cls, bytes_):
        struct_tuple = struct.unpack("<Q", bytes_)
        value = struct_tuple[0]
        return cls(value=value)


class Varuint32(AntelopeType):
    value: pydantic.conint(ge=0, le=20989371979)

    def __bytes__(self):
        bytes_ = b""
        val = self.value
        while True:
            b = val & 0x7F
            val >>= 7
            b |= (val > 0) << 7
            uint8 = Uint8(value=b)
            bytes_ += bytes(uint8)
            if not val:
                break
        return bytes_

    @classmethod
    def from_bytes(cls, bytes_):
        offset = 0
        value = 0
        for n, byte in enumerate(bytes_):
            partial_value = byte & 0x7F  # only the 7 first bits matter
            partial_value_offset = partial_value << offset
            value |= partial_value_offset
            offset += 7
            if n >= 8:
                break
            if not byte & 0x80:  # first bit (carry) off
                break
        return cls(value=value)


def _get_all_types():
    def is_eostype(class_):
        if isinstance(class_, type):
            if issubclass(class_, AntelopeType) and class_ is not AntelopeType:
                return True
        return False

    classes = list(sys.modules[__name__].__dict__.items())

    all_types = {
        name.lower(): class_ for name, class_ in classes if is_eostype(class_)
    }
    return all_types


_all_types = _get_all_types()


def from_string(type_: str) -> AntelopeType:
    type_ = type_.lower()
    try:
        class_ = _all_types[type_]
    except KeyError:
        types = list(_all_types.keys())
        msg = f"Type {type_} not found. List of available {types=}"
        raise ValueError(msg)
    return class_



"""Utility functions."""

import hashlib
import hmac
import io
import re
import struct

import base58

A = 0
P = 2**256 - 2**32 - 977
N = 115792089237316195423570985008687907852837564279074904382605163141518161494337  # NOQA: E501
Gx = 55066263022277343669578718895168534326250603453777594175500187360389116729240  # NOQA: E501
Gy = 32670510020758816978083085130507043184471273380659243275938904335757337482424  # NOQA: E501
G = (Gx, Gy)


def sign_bytes(*, bytes_: bytes, key: str) -> str:
    _check_bytes(bytes_)

    nonce = 0
    sha256 = hashlib.sha256()
    sha256.update(bytes_)
    while True:
        v, r, s = _ecdsa_raw_sign_nonce(sha256.digest(), key, nonce)
        signature = v.to_bytes(1, "big")
        signature += r.to_bytes(32, "big") + s.to_bytes(32, "big")
        if _is_canonical(signature):
            signature = b"\x00" + signature
            break
        nonce += 1

    ds = io.BytesIO(signature)
    first_byte = ds.read(1)
    t = struct.unpack("<B", first_byte)[0]
    if t == 0:
        data = ds.read(65)
        data = data + _ripmed160(data + b"K1")[:4]
        signature = "SIG_K1_" + base58.b58encode(data).decode("ascii")
    elif t == 1:
        raise NotImplementedError("Not Implemented")
    else:
        raise ValueError("Invalid binary signature")

    return signature


def _check_bytes(bytes_):
    if len(bytes_) == 0:
        raise ValueError("Can not sign empty bytes")
    if not isinstance(bytes_, bytes):
        msg = f"sign_bytes requires 'byte' type. But '{type(bytes_)}' received"
        raise TypeError(msg)


def _ripmed160(data):
    try:
        h = hashlib.new("ripemd160")
    except ValueError:
        from Crypto.Hash import RIPEMD160  # NOQA: I001

        h = RIPEMD160.new()
    h.update(data)
    return h.digest()


def _is_canonical(signature):
    canonical = all(
        [
            not (signature[1] & 0x80),
            not (signature[1] == 0 and not (signature[2] & 0x80)),
            not (signature[33] & 0x80),
            not (signature[33] == 0 and not (signature[34] & 0x80)),
        ]
    )
    return canonical


def _ecdsa_raw_sign_nonce(message_hash, key, nonce):
    msg_int = _decode(message_hash, 256)
    k = _deterministic_generate_k_nonce(message_hash, key, nonce)

    r, y = _fast_multiply(G, k)
    s = _inv(k, N) * (msg_int + r * _decode_privkey(key)) % N  # NOQA: W503

    v = 27 + ((y % 2) ^ (0 if s * 2 < N else 1))
    s = s if s * 2 < N else N - s

    if "compressed" in _get_privkey_format(key):
        v += 4

    return v, r, s


def _deterministic_generate_k_nonce(message_hash, key, nonce):
    v = b"\x01" * 32
    k = b"\x00" * 32
    try:
        key_encoded = _encode_privkey(key, "bin")
    except AssertionError:
        raise ValueError("Error in private key provided: {key=}")

    msg_int = _decode(message_hash, 256)
    message_hash = _encode(msg_int + nonce, 256, 32)

    k = _sha256_hmac_digest(k, v + b"\x00" + key_encoded + message_hash)
    v = _sha256_hmac_digest(k, v)
    k = _sha256_hmac_digest(k, v + b"\x01" + key_encoded + message_hash)
    v = _sha256_hmac_digest(k, v)
    k = _sha256_hmac_digest(k, v)

    k = _decode(k, 256)
    return k


def _sha256_hmac_digest(key, msg):
    value = hmac.new(key, msg, digestmod=hashlib.sha256)
    digest = value.digest()
    return digest


CODE_STRINGS = {
    2: "01",
    10: "0123456789",
    16: "0123456789abcdef",
    32: "abcdefghijklmnopqrstuvwxyz234567",
    58: "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz",
    256: "".join([chr(x) for x in range(256)]),
}


def _encode(val, base, minlen=0):
    base, minlen = int(base), int(minlen)
    code_string = CODE_STRINGS[base]
    result_bytes = bytes()
    while val > 0:
        curcode = code_string[val % base]
        result_bytes = bytes([ord(curcode)]) + result_bytes
        val //= base

    pad_size = minlen - len(result_bytes)

    if base == 256:
        padding_element = b"\x00"
    elif base == 58:
        padding_element = b"1"
    else:
        padding_element = b"0"

    if pad_size > 0:
        result_bytes = padding_element * pad_size + result_bytes

    result_string = "".join([chr(y) for y in result_bytes])
    result = result_bytes if base == 256 else result_string

    return result


def _decode(string, base):  # NOQA: C901
    base = int(base)
    code_string = CODE_STRINGS[base]
    result = 0
    if base == 256:

        def extract(d, cs):
            return d

    else:

        def extract(d, cs):
            return cs.find(d if isinstance(d, str) else chr(d))

    while len(string) > 0:
        result *= base
        result += extract(string[0], code_string)
        string = string[1:]
    return result


def _changebase(string, frm, to, minlen=0):
    if frm == to:
        char = CODE_STRINGS[frm][0]
        return string.ljust(minlen, char)
    return _encode(_decode(string, frm), to, minlen)


def _bin_dbl_sha256(s):
    bytes_to_hash = _from_string_to_bytes(s)
    return hashlib.sha256(hashlib.sha256(bytes_to_hash).digest()).digest()


def _from_string_to_bytes(a):
    return a if isinstance(a, bytes) else bytes(a, "utf-8")


def _b58check_to_bin(inp):
    leadingzbytes = len(re.match("^1*", inp).group(0))
    data = b"\x00" * leadingzbytes + _changebase(inp, 58, 256)
    assert _bin_dbl_sha256(data[:-4])[:4] == data[-4:]
    return data[1:-4]


def _get_privkey_format(priv):
    bin_p = _b58check_to_bin(priv)
    if len(bin_p) == 32:
        return "wif"
    else:
        msg = "Can't handle this private key format"
        raise NotImplementedError(msg)


def _from_int_to_byte(a):
    return bytes([a])


def _bin_to_b58check(inp, magicbyte=0):
    if magicbyte == 0:
        inp = _from_int_to_byte(0) + inp
    while magicbyte > 0:
        inp = _from_int_to_byte(magicbyte % 256) + inp
        magicbyte //= 256

    leadingzbytes = 0
    for x in inp:
        if x != 0:
            break
        leadingzbytes += 1

    checksum = _bin_dbl_sha256(inp)[:4]
    return "1" * leadingzbytes + _changebase(inp + checksum, 256, 58)


def _encode_privkey(priv, formt, vbyte=0):
    if not isinstance(priv, (int, float)):
        return _encode_privkey(_decode_privkey(priv), formt, vbyte)
    if formt == "bin":
        return _encode(priv, 256, 32)
    else:
        msg = "Can't handle this private key format"
        raise NotImplementedError(msg)


def _decode_privkey(priv, formt=None):
    if not formt:
        formt = _get_privkey_format(priv)
    if formt == "wif":
        return _decode(_b58check_to_bin(priv), 256)
    else:
        msg = "Can't handle this private key format"
        raise NotImplementedError(msg)


def _from_jacobian(p):
    z = _inv(p[2], P)
    return ((p[0] * z**2) % P, (p[1] * z**3) % P)


def _jacobian_double(p):
    if not p[1]:
        return (0, 0, 0)
    ysq = (p[1] ** 2) % P
    S = (4 * p[0] * ysq) % P
    M = (3 * p[0] ** 2 + A * p[2] ** 4) % P
    nx = (M**2 - 2 * S) % P
    ny = (M * (S - nx) - 8 * ysq**2) % P
    nz = (2 * p[1] * p[2]) % P
    return (nx, ny, nz)


def _jacobian_multiply(a, n):  # NOQA: C901
    if a[1] == 0 or n == 0:
        return (0, 0, 1)
    if n == 1:
        return a
    if n < 0 or n >= N:
        return _jacobian_multiply(a, n % N)
    if (n % 2) == 0:
        return _jacobian_double(_jacobian_multiply(a, n // 2))
    if (n % 2) == 1:
        return _jacobian_add(
            _jacobian_double(_jacobian_multiply(a, n // 2)), a
        )


def _jacobian_add(p, q):
    if not p[1]:
        return q
    if not q[1]:
        return p
    U1 = (p[0] * q[2] ** 2) % P
    U2 = (q[0] * p[2] ** 2) % P
    S1 = (p[1] * q[2] ** 3) % P
    S2 = (q[1] * p[2] ** 3) % P
    if U1 == U2:
        if S1 != S2:
            return (0, 0, 1)
        return _jacobian_double(p)
    H = U2 - U1
    R = S2 - S1
    H2 = (H * H) % P
    H3 = (H * H2) % P
    U1H2 = (U1 * H2) % P
    nx = (R**2 - H3 - 2 * U1H2) % P
    ny = (R * (U1H2 - nx) - S1 * H3) % P
    nz = (H * p[2] * q[2]) % P
    return (nx, ny, nz)


def _to_jacobian(p):
    o = (p[0], p[1], 1)
    return o


def _fast_multiply(a, n):
    return _from_jacobian(_jacobian_multiply(_to_jacobian(a), n))


def _inv(a, n):
    if a == 0:
        return 0
    lm, hm = 1, 0
    low, high = a % n, n
    while low > 1:
        r = high // low
        nm, new = hm - lm * r, high - low * r
        lm, low, hm, high = nm, new, lm, low
    return lm % n



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyntelope']

package_data = \
{'': ['*']}

install_requires = \
['base58>=2.1.1,<3.0.0',
 'httpx>=0.22',
 'pycryptodome>=3.15.0,<4.0.0',
 'pydantic>=1.9.0,<2.0.0']

setup_kwargs = {
    'name': 'pyntelope',
    'version': '0.6.4',
    'description': 'Interact with Antelope blockchains',
    'long_description': '<div align="center">\n    \n<p align="center">\n  <img width="600" src="https://miro.medium.com/max/1400/1*5KEvJB1UBBsk_1ZTBtJfJA.png">\n</p>\n    \n*Minimalist python library to interact with antelope blockchain networks*\n \n![Test](https://github.com/FACINGS/pyntelope/actions/workflows/main_workflow.yml/badge.svg)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/pyntelope)\n![version](https://img.shields.io/pypi/v/pyntelope)\n![GitHub repo size](https://img.shields.io/github/repo-size/facings/pyntelope)\n![GitHub last commit](https://img.shields.io/github/last-commit/facings/pyntelope)\n\n</div>\n\n# What is it?\n**pyntelope** is a python library to interact with Antelope blockchains.  \nIts main focus are server side applications.  \nThis library is heavily influenced by [µEOSIO](https://github.com/EOSArgentina/ueosio). Many thanks to them for the astonishing job!  \n\n\n# Main features\n- Send transactions\nIts main usage today is to send transactions to the blockchain\n- Statically typed\nThis library enforces and verifies types and values.\n- Serialization\n**pyntelope** serializes the transaction before sending to the blockchain. \n- Paralellization\nAlthough python has the [GIL](https://realpython.com/python-gil/) we try to make as easier as possible to paralellize the jobs.  \nAll data is as immutable and all functions are as pure as we can make them.  \n\n\n# Stability\nThis work is in alpha version. That means that we make constant breaking changes to its api.  \nAlso there are known (and, of course unknown) bugs and various limitations.  \nGiven that, we at [FACINGS](https://facings.io/) have been using this library in production for over an year now.  \nHowever we\'d advise for you to fix its version when deploying to prod.  \n\n\n# Using\nJust `pip install pyntelope` and play around.  \n(we don\'t support, and have no plans to support [conda](https://docs.conda.io/en/latest/))  \nRather then starting with long docs, just a simple example:  \n\n\n## Use Send Message action\n```python\nimport pyntelope\n\n\nprint("Create Transaction")\ndata=[\n    pyntelope.Data(\n        name="from",\n        value=pyntelope.types.Name("me.wam"), \n    ),\n    pyntelope.Data(\n        name="message",\n         value=pyntelope.types.String("hello from pyntelope"),\n    ),\n]\n\nauth = pyntelope.Authorization(actor="me.wam", permission="active")\n\naction = pyntelope.Action(\n    account="me.wam", # this is the contract account\n    name="sendmsg", # this is the action name\n    data=data,\n    authorization=[auth],\n)\n\nraw_transaction = pyntelope.Transaction(actions=[action])\n\nprint("Link transaction to the network")\nnet = pyntelope.WaxTestnet()  # this is an alias for a testnet node\n# notice that pyntelope returns a new object instead of change in place\nlinked_transaction = raw_transaction.link(net=net)\n\n\nprint("Sign transaction")\nkey = "a_very_secret_key"\nsigned_transaction = linked_transaction.sign(key=key)\n\n\nprint("Send")\nresp = signed_transaction.send()\n\nprint("Printing the response")\nresp_fmt = json.dumps(resp, indent=4)\nprint(f"Response:\\n{resp_fmt}")\n```\n\nThere are some other examples [here](./examples)\n\n\n# Known bugs\n### multi-byte utf-8 characters can not be serialized\n- Serialization of multi-byte utf-8 characters is somewhat unpredictable in the current implementation, therfore any String input containing multi-utf8 byte characters will be blocked for the time being.\n\n\n# Contributing\nAll contributions, bug reports, bug fixes, documentation improvements, enhancements, and ideas are welcome.  \nIf you find a bug, just open a issue with a tag "BUG".  \nIf you want to request a new feature, open an issue with a tag "ENH" (for enhancement).  \nIf you feel like that our docs could be better, please open one with a tag "DOC".  \nAlthough we have the next few steps already planned, we are happy to receive the community feedback to see where to go from there.  \n\n\n### Development\nIf you want to develop for **pyntelope**, here are some tips for a local development environment.\nWe\'ll be more then happy to receive PRs from the community.\nAlso we\'re going full [Black](https://black.readthedocs.io/en/stable/) and enforcing [pydocstyle](http://www.pydocstyle.org/en/stable/) and [isort](https://pypi.org/project/isort/) (with the limitations described in the .flake8 file)\n\n#### Setup\nCreate a virtual env\nEnsure the dependencies are met:\n```\npip install poetry\npoetry install\n```\n\n#### Run tests\nThe tests are run against a local network.  \nBefore running the tests you\'ll need to `docker-compose up` to create the local network, users and contracts used in the tests.  \nWhen ready, just:\n```\npytest\n```\n\n\n\n',
    'author': 'Team',
    'author_email': 'pyntelope@facings.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/FACINGS/pyntelope',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Transaction, Authorization and Action classes."""


import datetime as dt
import hashlib
import json
import struct
from typing import List, Tuple

import pydantic

from . import types, utils
from .net import Net


class Authorization(pydantic.BaseModel):
    """
    Authorization to be used in Action.

    actor: str
    permission: str
    """

    actor: pydantic.constr(min_length=1, max_length=13)
    permission: pydantic.constr(min_length=1, max_length=13)

    def __bytes__(self):
        bytes_ = b""
        account_name = types.Name(self.actor)
        bytes_ += bytes(account_name)
        permission = types.Name(self.permission)
        bytes_ += bytes(permission)
        return bytes_

    class Config:
        extra = "forbid"
        frozen = True


class Data(pydantic.BaseModel):
    """
    Data to be used in actions.

    name: the data field name
    value: the typed value (types.AntelopeType) of the data
    """

    name: str
    value: types.AntelopeType

    def __init__(self, *args, **kwargs):
        if len(args) == 1:
            if isinstance(args[0], dict):
                self = self.parse_obj(args[0])
                return
        super().__init__(*args, **kwargs)

    @classmethod
    def parse_obj(self, obj):
        for field in ["name", "type", "value"]:
            if field not in obj:
                msg = f"Field {field} expected. {obj}"
                raise ValueError(msg)
        if len(obj) != 3:
            msg = (
                f"Object with lenght 3 was expected, but {len(obj)} "
                f"found: {obj}"
            )
            raise ValueError(msg)
        name = obj["name"]
        type_str = obj["type"]
        value_raw = obj["value"]
        type_obj = types.from_string(type_str)
        value = type_obj(value_raw)
        return Data(name=name, value=value)

    def dict(self):
        d = dict(
            name=self.name,
            type=self.value.__class__.__name__,
            value=self.value.value,
        )
        return d

    def json(self):
        d = self.dict()
        j = json.dumps(d)
        return j

    def __bytes__(self):
        return bytes(self.value)

    class Config:
        extra = "forbid"
        frozen = True


class Action(pydantic.BaseModel):
    """
    Action to be used in Transaction.

    account: str
    name: str
    data: list[Data]
    authorization: list[Action]
    """

    account: pydantic.constr(max_length=13)
    name: str
    authorization: pydantic.conlist(Authorization, min_items=1, max_items=10)
    data: List[Data]

    @pydantic.validator("data", "authorization")
    def transform_to_tuple(cls, v):
        new_v = tuple(v)
        return new_v

    # returns a LinkedAction with current values and a specificed net value
    def link(self, net: Net):
        return LinkedAction(
            account=self.account,
            name=self.name,
            authorization=self.authorization,
            data=self.data,
            net=net,
        )

    def __bytes__(self):
        name = self.__class__.__name__
        raise TypeError(f"cannot convert '{name}' object to bytes")

    class Config:
        extra = "forbid"
        frozen = True
        arbitrary_types_allowed = True


class LinkedAction(Action):
    """
    Action to be used in LinkedTransaction.

    account: str
    name: str
    data: list[Data]
    authorization: list[Authorization]
    """

    account: pydantic.constr(max_length=13)
    name: str
    authorization: pydantic.conlist(Authorization, min_items=1, max_items=10)
    data: List[Data]
    net: Net

    def __bytes__(self):
        bytes_ = b""
        account_name = types.Name(value=self.account)
        bytes_ += bytes(account_name)
        action_name = types.Name(value=self.name)
        bytes_ += bytes(action_name)

        auth_bytes = [bytes(a) for a in self.authorization]
        auth = types.Array(type_=types.Bytes, values=auth_bytes)
        bytes_ += bytes(auth)

        data_bytes = b""
        for d in self.data:
            data_bytes += bytes(d)
        data_bytes = data_bytes.hex()
        data_bytes_list = []
        for i in range(0, len(data_bytes), 2):
            data_bytes_list.append(data_bytes[i : i + 2])  # NOQA: E203
        data_bytes_list = [bytes.fromhex(b) for b in data_bytes_list]
        data = types.Array(type_=types.Bytes, values=data_bytes_list)

        bytes_ += bytes(data)

        return bytes_


def _endian_reverse_u32(i: int) -> int:
    i = i & 0xFFFFFFFF
    r = (((i >> 0x18) & 0xFF)) | (((i >> 0x10) & 0xFF) << 0x08) | (((i >> 0x08) & 0xFF) << 0x10) | (((i) & 0xFF) << 0x18)  # NOQA BLK100, E501
    return r


def _get_tapos_info(block_id: str) -> Tuple[int]:
    block_id_bin = bytes.fromhex(block_id)

    hash0 = struct.unpack("<Q", block_id_bin[0:8])[0]
    hash1 = struct.unpack("<Q", block_id_bin[8:16])[0]

    ref_block_num = _endian_reverse_u32(hash0) & 0xFFFF
    ref_block_prefix = hash1 & 0xFFFFFFFF
    return ref_block_num, ref_block_prefix


class Transaction(pydantic.BaseModel):
    """
    Raw Transaction. It can't be sent to the blockchain.

    It becomes a LinkedTransaction when a Net is linked

    actions: list[Action]
    delay_sec: int = 0
    max_cpu_usage_ms: int = 0
    chain_id: Optional[str]
    """

    actions: pydantic.conlist(Action, min_items=1, max_items=10)
    expiration_delay_sec: pydantic.conint(ge=0) = 600
    delay_sec: pydantic.conint(ge=0) = 0
    max_cpu_usage_ms: pydantic.conint(ge=0) = 0
    max_net_usage_words: pydantic.conint(ge=0) = 0

    @pydantic.validator("actions")
    def _transform_to_tuple(cls, v):
        new_v = tuple(v)
        return new_v

    # used to link transaction to a specified network (net)
    # gets required info from net then returns a LinkedTransaction
    def link(self, *, net: Net):  # block_id: str, chain_id: str):
        net_info = net.get_info()
        block_id = net_info["last_irreversible_block_id"]
        chain_id = net_info["chain_id"]

        ref_block_num, ref_block_prefix = _get_tapos_info(block_id=block_id)
        expiration = dt.datetime.utcnow() + dt.timedelta(
            seconds=self.expiration_delay_sec
        )

        new_trans = LinkedTransaction(
            # load every action as a linkedAction with the net passed in
            actions=[a.link(net) for a in self.actions],
            net=net,
            expiration_delay_sec=self.expiration_delay_sec,
            delay_sec=self.delay_sec,
            max_cpu_usage_ms=self.max_cpu_usage_ms,
            max_net_usage_words=self.max_net_usage_words,
            chain_id=chain_id,
            ref_block_num=ref_block_num,
            ref_block_prefix=ref_block_prefix,
            expiration=expiration,
        )

        return new_trans

    class Config:
        extra = "forbid"
        frozen = True
        arbitrary_types_allowed = True


class LinkedTransaction(Transaction):
    """
    Linked transaction. It can't be sent to the blockchain.

    It becomes a SignedTransaction when you sign it.
    """

    actions: pydantic.conlist(LinkedAction, min_items=1, max_items=10)
    net: Net
    chain_id: str
    ref_block_num: str
    ref_block_prefix: str
    expiration: dt.datetime

    def __bytes__(self):
        bytes_ = b""
        bytes_ += bytes(types.UnixTimestamp(self.expiration))
        bytes_ += bytes(types.Uint16(self.ref_block_num))
        bytes_ += bytes(types.Uint32(self.ref_block_prefix))
        bytes_ += bytes(types.Varuint32(self.max_net_usage_words))
        bytes_ += bytes(types.Uint8(self.max_cpu_usage_ms))
        bytes_ += bytes(types.Varuint32(self.delay_sec))
        # context_free_actions
        bytes_ += bytes(types.Array(type_=types.Int8, values=[]))

        actions_bytes = [bytes(act) for act in self.actions]
        actions = types.Array(type_=types.Bytes, values=actions_bytes)
        bytes_ += bytes(actions)

        # transaction_extensions
        bytes_ += bytes(types.Array(type_=types.Int8, values=[]))

        return bytes_

    def id(self):
        hash256 = hashlib.sha256()
        hash256.update(bytes(self))
        hash256_digest = hash256.digest()
        return hash256_digest.hex()

    def sign(self, key: str):
        signs = []
        if hasattr(self, "signatures"):
            signs = list(self.signatures)

        chain_bytes = bytes.fromhex(self.chain_id)
        trans_bytes = bytes(self)
        zero_bytes = bytes.fromhex("0" * 64)
        bytes_ = chain_bytes + trans_bytes + zero_bytes

        signature = utils.sign_bytes(bytes_=bytes_, key=key)
        signs.append(signature)
        trans = SignedTransaction(
            net=self.net,
            actions=self.actions,
            expiration_delay_sec=self.expiration_delay_sec,
            delay_sec=self.delay_sec,
            max_cpu_usage_ms=self.max_cpu_usage_ms,
            max_net_usage_words=self.max_net_usage_words,
            chain_id=self.chain_id,
            ref_block_num=self.ref_block_num,
            ref_block_prefix=self.ref_block_prefix,
            expiration=self.expiration,
            signatures=tuple(signs),
        )
        return trans


class SignedTransaction(LinkedTransaction):
    """
    Signed transaction. You can send it to the blockchain.

    Also you can sign it again.
    """

    signatures: pydantic.conlist(str, min_items=1, max_items=10)

    @pydantic.validator("signatures")
    def _transform_to_tuple(cls, v):
        new_v = tuple(v)
        return new_v

    def pack(self):
        bytes_ = bytes(self)
        return bytes_.hex()

    def send(self):
        resp = self.net.push_transaction(transaction=self)
        return resp


__all__ = [
    "Action",
    "Authorization",
    "Data",
    "Transaction",
    "LinkedTransaction",
    "SignedTransaction",
    "LinkedAction",
]



"""Exceptions."""

import logging

logger = logging.getLogger(__name__)


class ConnectionError(Exception):
    def __init__(self, *, response, url, payload, error):
        try:
            text = response.text
        except AttributeError:
            text = ""
        msg = [
            f"{response=}",
            f"{text=}",
            f"{url=}",
            f"{payload=}",
            f"{error=}",
        ]
        msg = ";\n".join(msg)
        super().__init__(self, msg)
